@@ -27,16 +27,26 @@
 
 #include "code/oopRecorder.hpp"
 #include "code/relocInfo.hpp"
+#include "compiler/compiler_globals.hpp"
 #include "utilities/align.hpp"
 #include "utilities/debug.hpp"
+#include "utilities/growableArray.hpp"
+#include "utilities/linkedlist.hpp"
+#include "utilities/resizeableResourceHash.hpp"
 #include "utilities/macros.hpp"
 
-class CodeStrings;
+template <typename T>
+static inline void put_native(address p, T x) {
+    memcpy((void*)p, &x, sizeof x);
+}
+
 class PhaseCFG;
 class Compile;
 class BufferBlob;
 class CodeBuffer;
 class Label;
+class ciMethod;
+class SharedStubToInterpRequest;
 
 class CodeOffsets: public StackObj {
 public:
@@ -93,35 +103,37 @@ class CodeSection {
   address     _locs_point;      // last relocated position (grows upward)
   bool        _locs_own;        // did I allocate the locs myself?
   bool        _scratch_emit;    // Buffer is used for scratch emit, don't relocate.
-  char        _index;           // my section number (SECT_INST, etc.)
+  int         _skipped_instructions_size;
+  int8_t      _index;           // my section number (SECT_INST, etc.)
   CodeBuffer* _outer;           // enclosing CodeBuffer
 
   // (Note:  _locs_point used to be called _last_reloc_offset.)
 
   CodeSection() {
-    _start         = NULL;
-    _mark          = NULL;
-    _end           = NULL;
-    _limit         = NULL;
-    _locs_start    = NULL;
-    _locs_end      = NULL;
-    _locs_limit    = NULL;
-    _locs_point    = NULL;
+    _start         = nullptr;
+    _mark          = nullptr;
+    _end           = nullptr;
+    _limit         = nullptr;
+    _locs_start    = nullptr;
+    _locs_end      = nullptr;
+    _locs_limit    = nullptr;
+    _locs_point    = nullptr;
     _locs_own      = false;
     _scratch_emit  = false;
-    debug_only(_index = (char)-1);
+    _skipped_instructions_size = 0;
+    debug_only(_index = -1);
     debug_only(_outer = (CodeBuffer*)badAddress);
   }
 
-  void initialize_outer(CodeBuffer* outer, int index) {
+  void initialize_outer(CodeBuffer* outer, int8_t index) {
     _outer = outer;
     _index = index;
   }
 
   void initialize(address start, csize_t size = 0) {
-    assert(_start == NULL, "only one init step, please");
+    assert(_start == nullptr, "only one init step, please");
     _start         = start;
-    _mark          = NULL;
+    _mark          = nullptr;
     _end           = start;
 
     _limit         = start + size;
@@ -139,6 +151,7 @@ class CodeSection {
     _end        = cs->_end;
     _limit      = cs->_limit;
     _locs_point = cs->_locs_point;
+    _skipped_instructions_size = cs->_skipped_instructions_size;
   }
 
  public:
@@ -147,7 +160,7 @@ class CodeSection {
   address     end() const           { return _end; }
   address     limit() const         { return _limit; }
   csize_t     size() const          { return (csize_t)(_end - _start); }
-  csize_t     mark_off() const      { assert(_mark != NULL, "not an offset");
+  csize_t     mark_off() const      { assert(_mark != nullptr, "not an offset");
                                       return (csize_t)(_mark - _start); }
   csize_t     capacity() const      { return (csize_t)(_limit - _start); }
   csize_t     remaining() const     { return (csize_t)(_limit - _end); }
@@ -160,13 +173,14 @@ class CodeSection {
   csize_t     locs_point_off() const{ return (csize_t)(_locs_point - _start); }
   csize_t     locs_capacity() const { return (csize_t)(_locs_limit - _locs_start); }
 
-  int         index() const         { return _index; }
-  bool        is_allocated() const  { return _start != NULL; }
+  int8_t      index() const         { return _index; }
+  bool        is_allocated() const  { return _start != nullptr; }
   bool        is_empty() const      { return _start == _end; }
-  bool        has_locs() const      { return _locs_end != NULL; }
+  bool        has_locs() const      { return _locs_end != nullptr; }
 
   // Mark scratch buffer.
   void        set_scratch_emit()    { _scratch_emit = true; }
+  void        clear_scratch_emit()  { _scratch_emit = false; }
   bool        scratch_emit()        { return _scratch_emit; }
 
   CodeBuffer* outer() const         { return _outer; }
@@ -187,7 +201,7 @@ class CodeSection {
   void    set_mark(address pc)      { assert(contains2(pc), "not in codeBuffer");
                                       _mark = pc; }
   void    set_mark()                { _mark = _end; }
-  void    clear_mark()              { _mark = NULL; }
+  void    clear_mark()              { _mark = nullptr; }
 
   void    set_locs_end(relocInfo* p) {
     assert(p <= locs_limit(), "locs data fits in allocated buffer");
@@ -199,48 +213,50 @@ class CodeSection {
     _locs_point = pc;
   }
 
+  void register_skipped(int size) {
+    _skipped_instructions_size += size;
+  }
+
   // Code emission
-  void emit_int8(int8_t x1) {
+  void emit_int8(uint8_t x1) {
     address curr = end();
-    *((int8_t*)  curr++) = x1;
+    *((uint8_t*)  curr++) = x1;
     set_end(curr);
   }
 
-  void emit_int16(int16_t x) { *((int16_t*) end()) = x; set_end(end() + sizeof(int16_t)); }
-  void emit_int16(int8_t x1, int8_t x2) {
+  template <typename T>
+  void emit_native(T x) { put_native(end(), x); set_end(end() + sizeof x); }
+
+  void emit_int16(uint16_t x) { emit_native(x); }
+  void emit_int16(uint8_t x1, uint8_t x2) {
     address curr = end();
-    *((int8_t*)  curr++) = x1;
-    *((int8_t*)  curr++) = x2;
+    *((uint8_t*)  curr++) = x1;
+    *((uint8_t*)  curr++) = x2;
     set_end(curr);
   }
 
-  void emit_int24(int8_t x1, int8_t x2, int8_t x3)  {
+  void emit_int24(uint8_t x1, uint8_t x2, uint8_t x3)  {
     address curr = end();
-    *((int8_t*)  curr++) = x1;
-    *((int8_t*)  curr++) = x2;
-    *((int8_t*)  curr++) = x3;
+    *((uint8_t*)  curr++) = x1;
+    *((uint8_t*)  curr++) = x2;
+    *((uint8_t*)  curr++) = x3;
     set_end(curr);
   }
 
-  void emit_int32(int32_t x) {
+  void emit_int32(uint32_t x) { emit_native(x); }
+  void emit_int32(uint8_t x1, uint8_t x2, uint8_t x3, uint8_t x4)  {
     address curr = end();
-    *((int32_t*) curr) = x;
-    set_end(curr + sizeof(int32_t));
-  }
-  void emit_int32(int8_t x1, int8_t x2, int8_t x3, int8_t x4)  {
-    address curr = end();
-    *((int8_t*)  curr++) = x1;
-    *((int8_t*)  curr++) = x2;
-    *((int8_t*)  curr++) = x3;
-    *((int8_t*)  curr++) = x4;
+    *((uint8_t*)  curr++) = x1;
+    *((uint8_t*)  curr++) = x2;
+    *((uint8_t*)  curr++) = x3;
+    *((uint8_t*)  curr++) = x4;
     set_end(curr);
   }
 
-  void emit_int64( int64_t x)  { *((int64_t*) end()) = x; set_end(end() + sizeof(int64_t)); }
-
-  void emit_float( jfloat  x)  { *((jfloat*)  end()) = x; set_end(end() + sizeof(jfloat)); }
-  void emit_double(jdouble x)  { *((jdouble*) end()) = x; set_end(end() + sizeof(jdouble)); }
-  void emit_address(address x) { *((address*) end()) = x; set_end(end() + sizeof(address)); }
+  void emit_int64(uint64_t x)  { emit_native(x); }
+  void emit_float(jfloat  x)   { emit_native(x); }
+  void emit_double(jdouble x)  { emit_native(x); }
+  void emit_address(address x) { emit_native(x); }
 
   // Share a scratch buffer for relocinfo.  (Hacky; saves a resource allocation.)
   void initialize_shared_locs(relocInfo* buf, int length);
@@ -252,16 +268,14 @@ class CodeSection {
   void relocate(address at, RelocationHolder const& rspec, int format = 0);
   void relocate(address at,    relocInfo::relocType rtype, int format = 0, jint method_index = 0);
 
-  // alignment requirement for starting offset
-  // Requirements are that the instruction area and the
-  // stubs area must start on CodeEntryAlignment, and
-  // the ctable on sizeof(jdouble)
-  int alignment() const             { return MAX2((int)sizeof(jdouble), (int)CodeEntryAlignment); }
+  int alignment() const;
 
   // Slop between sections, used only when allocating temporary BufferBlob buffers.
   static csize_t end_slop()         { return MAX2((int)sizeof(jdouble), (int)CodeEntryAlignment); }
 
-  csize_t align_at_start(csize_t off) const { return (csize_t) align_up(off, alignment()); }
+  csize_t align_at_start(csize_t off) const {
+    return (csize_t) align_up(off, alignment());
+  }
 
   // Ensure there's enough space left in the current section.
   // Return true if there was an expansion.
@@ -273,71 +287,78 @@ class CodeSection {
 #endif //PRODUCT
 };
 
-class CodeString;
-class CodeStrings {
+
+#ifndef PRODUCT
+
+class AsmRemarkCollection;
+class DbgStringCollection;
+
+// The assumption made here is that most code remarks (or comments) added to
+// the generated assembly code are unique, i.e. there is very little gain in
+// trying to share the strings between the different offsets tracked in a
+// buffer (or blob).
+
+class AsmRemarks {
+ public:
+  AsmRemarks();
+ ~AsmRemarks();
+
+  const char* insert(uint offset, const char* remstr);
+
+  bool is_empty() const;
+
+  void share(const AsmRemarks &src);
+  void clear();
+  uint print(uint offset, outputStream* strm = tty) const;
+
+  // For testing purposes only.
+  const AsmRemarkCollection* ref() const { return _remarks; }
+
 private:
-#ifndef PRODUCT
-  CodeString* _strings;
-  CodeString* _strings_last;
-#ifdef ASSERT
-  // Becomes true after copy-out, forbids further use.
-  bool _defunct; // Zero bit pattern is "valid", see memset call in decode_env::decode_env
-#endif
-  static const char* _prefix; // defaults to " ;; "
-
-  CodeString* find(intptr_t offset) const;
-  CodeString* find_last(intptr_t offset) const;
-
-  void set_null_and_invalidate() {
-    _strings = NULL;
-    _strings_last = NULL;
-#ifdef ASSERT
-    _defunct = true;
-#endif
-  }
-#endif
-
-public:
-  CodeStrings() {
-#ifndef PRODUCT
-    _strings = NULL;
-    _strings_last = NULL;
-#ifdef ASSERT
-    _defunct = false;
-#endif
-#endif
-  }
-
-#ifndef PRODUCT
-  bool is_null() {
-#ifdef ASSERT
-    return _strings == NULL;
-#else
-    return true;
-#endif
-  }
-
-  const char* add_string(const char * string);
-
-  void add_comment(intptr_t offset, const char * comment);
-  void print_block_comment(outputStream* stream, intptr_t offset) const;
-  int  count() const;
-  // COPY strings from other to this; leave other valid.
-  void copy(CodeStrings& other);
-  // FREE strings; invalidate this.
-  void free();
-
-  // Guarantee that _strings are used at most once; assign and free invalidate a buffer.
-  inline void check_valid() const {
-    assert(!_defunct, "Use of invalid CodeStrings");
-  }
-
-  static void set_prefix(const char *prefix) {
-    _prefix = prefix;
-  }
-#endif // !PRODUCT
+  AsmRemarkCollection* _remarks;
 };
 
+// The assumption made here is that the number of debug strings (with a fixed
+// address requirement) is a rather small set per compilation unit.
+
+class DbgStrings {
+ public:
+  DbgStrings();
+ ~DbgStrings();
+
+  const char* insert(const char* dbgstr);
+
+  bool is_empty() const;
+
+  void share(const DbgStrings &src);
+  void clear();
+
+  // For testing purposes only.
+  const DbgStringCollection* ref() const { return _strings; }
+
+private:
+  DbgStringCollection* _strings;
+};
+#endif // not PRODUCT
+
+
+#ifdef ASSERT
+#include "utilities/copy.hpp"
+
+class Scrubber {
+ public:
+  Scrubber(void* addr, size_t size) : _addr(addr), _size(size) {}
+ ~Scrubber() {
+    Copy::fill_to_bytes(_addr, _size, badResourceValue);
+  }
+ private:
+  void*  _addr;
+  size_t _size;
+};
+#endif // ASSERT
+
+typedef GrowableArray<SharedStubToInterpRequest> SharedStubToInterpRequests;
+
 // A CodeBuffer describes a memory space into which assembly
 // code is generated.  This memory space usually occupies the
 // interior of a single BufferBlob, but in some cases it may be
@@ -362,7 +383,7 @@ public:
 // Instructions and data in one section can contain relocatable references to
 // addresses in a sibling section.
 
-class CodeBuffer: public StackObj {
+class CodeBuffer: public StackObj DEBUG_ONLY(COMMA private Scrubber) {
   friend class CodeSection;
   friend class StubCodeGenerator;
 
@@ -370,12 +391,12 @@ class CodeBuffer: public StackObj {
   // CodeBuffers must be allocated on the stack except for a single
   // special case during expansion which is handled internally.  This
   // is done to guarantee proper cleanup of resources.
-  void* operator new(size_t size) throw() { return ResourceObj::operator new(size); }
+  void* operator new(size_t size) throw() { return resource_allocate_bytes(size); }
   void  operator delete(void* p)          { ShouldNotCallThis(); }
 
  public:
   typedef int csize_t;  // code size type; would be size_t except for history
-  enum {
+  enum : int8_t {
     // Here is the list of all possible sections.  The order reflects
     // the final layout.
     SECT_FIRST = 0,
@@ -385,6 +406,9 @@ class CodeBuffer: public StackObj {
     SECT_LIMIT, SECT_NONE = -1
   };
 
+  typedef LinkedListImpl<int> Offsets;
+  typedef ResizeableResourceHashtable<address, Offsets, AnyObj::C_HEAP, mtCompiler> SharedTrampolineRequests;
+
  private:
   enum {
     sect_bits = 2,      // assert (SECT_LIMIT <= (1<<sect_bits))
@@ -409,9 +433,17 @@ class CodeBuffer: public StackObj {
   Arena*       _overflow_arena;
 
   address      _last_insn;      // used to merge consecutive memory barriers, loads or stores.
+  address      _last_label;     // record last bind label address, it's also the start of current bb.
+
+  SharedStubToInterpRequests* _shared_stub_to_interp_requests; // used to collect requests for shared iterpreter stubs
+  SharedTrampolineRequests*   _shared_trampoline_requests;     // used to collect requests for shared trampolines
+  bool         _finalize_stubs; // Indicate if we need to finalize stubs to make CodeBuffer final.
+
+  int          _const_section_alignment;
 
 #ifndef PRODUCT
-  CodeStrings  _code_strings;
+  AsmRemarks   _asm_remarks;
+  DbgStrings   _dbg_strings;
   bool         _collect_comments; // Indicate if we need to collect block comments at all.
   address      _decode_begin;     // start address for decode
   address      decode_begin();
@@ -419,17 +451,28 @@ class CodeBuffer: public StackObj {
 
   void initialize_misc(const char * name) {
     // all pointers other than code_start/end and those inside the sections
-    assert(name != NULL, "must have a name");
+    assert(name != nullptr, "must have a name");
     _name            = name;
-    _before_expand   = NULL;
-    _blob            = NULL;
-    _oop_recorder    = NULL;
-    _overflow_arena  = NULL;
-    _last_insn       = NULL;
+    _before_expand   = nullptr;
+    _blob            = nullptr;
+    _oop_recorder    = nullptr;
+    _overflow_arena  = nullptr;
+    _last_insn       = nullptr;
+    _last_label      = nullptr;
+    _finalize_stubs  = false;
+    _shared_stub_to_interp_requests = nullptr;
+    _shared_trampoline_requests = nullptr;
+
+    _consts.initialize_outer(this, SECT_CONSTS);
+    _insts.initialize_outer(this,  SECT_INSTS);
+    _stubs.initialize_outer(this,  SECT_STUBS);
+
+    // Default is to align on 8 bytes. A compiler can change this
+    // if larger alignment (e.g., 32-byte vector masks) is required.
+    _const_section_alignment = (int) sizeof(jdouble);
 
 #ifndef PRODUCT
-    _decode_begin    = NULL;
-    _code_strings    = CodeStrings();
+    _decode_begin    = nullptr;
     // Collect block comments, but restrict collection to cases where a disassembly is output.
     _collect_comments = ( PrintAssembly
                        || PrintStubCode
@@ -442,9 +485,6 @@ class CodeBuffer: public StackObj {
   }
 
   void initialize(address code_start, csize_t code_size) {
-    _consts.initialize_outer(this,  SECT_CONSTS);
-    _insts.initialize_outer(this,   SECT_INSTS);
-    _stubs.initialize_outer(this,   SECT_STUBS);
     _total_start = code_start;
     _total_size  = code_size;
     // Initialize the main section:
@@ -472,6 +512,9 @@ class CodeBuffer: public StackObj {
   // moves code sections to new buffer (assumes relocs are already in there)
   void relocate_code_to(CodeBuffer* cb) const;
 
+  // adjust some internal address during expand
+  void adjust_internal_address(address from, address to);
+
   // set up a model of the final layout of my contents
   void compute_final_layout(CodeBuffer* dest) const;
 
@@ -484,8 +527,10 @@ class CodeBuffer: public StackObj {
 
  public:
   // (1) code buffer referring to pre-allocated instruction memory
-  CodeBuffer(address code_start, csize_t code_size) {
-    assert(code_start != NULL, "sanity");
+  CodeBuffer(address code_start, csize_t code_size)
+    DEBUG_ONLY(: Scrubber(this, sizeof(*this)))
+  {
+    assert(code_start != nullptr, "sanity");
     initialize_misc("static buffer");
     initialize(code_start, code_size);
     debug_only(verify_section_allocation();)
@@ -497,14 +542,18 @@ class CodeBuffer: public StackObj {
   // (3) code buffer allocating codeBlob memory for code & relocation
   // info but with lazy initialization.  The name must be something
   // informative.
-  CodeBuffer(const char* name) {
+  CodeBuffer(const char* name)
+    DEBUG_ONLY(: Scrubber(this, sizeof(*this)))
+  {
     initialize_misc(name);
   }
 
   // (4) code buffer allocating codeBlob memory for code & relocation
   // info.  The name must be something informative and code_size must
   // include both code and stubs sizes.
-  CodeBuffer(const char* name, csize_t code_size, csize_t locs_size) {
+  CodeBuffer(const char* name, csize_t code_size, csize_t locs_size)
+    DEBUG_ONLY(: Scrubber(this, sizeof(*this)))
+  {
     initialize_misc(name);
     initialize(code_size, locs_size);
   }
@@ -523,7 +572,7 @@ class CodeBuffer: public StackObj {
 
   const CodeSection* insts() const { return &_insts; }
 
-  // present sections in order; return NULL at end; consts is #0, etc.
+  // present sections in order; return null at end; consts is #0, etc.
   CodeSection* code_section(int n) {
     // This makes the slightly questionable but portable assumption
     // that the various members (_consts, _insts, _stubs, etc.) are
@@ -548,7 +597,7 @@ class CodeBuffer: public StackObj {
   static int locator(int pos, int sect) { return (pos << sect_bits) | sect; }
   int        locator(address addr) const;
   address    locator_address(int locator) const {
-    if (locator < 0)  return NULL;
+    if (locator < 0)  return nullptr;
     address start = code_section(locator_sect(locator))->start();
     return start + locator_pos(locator);
   }
@@ -605,18 +654,20 @@ class CodeBuffer: public StackObj {
   // allocated size of all relocation data, including index, rounded up
   csize_t total_relocation_size() const;
 
+  int total_skipped_instructions_size() const;
+
   csize_t copy_relocations_to(address buf, csize_t buf_limit, bool only_inst) const;
 
   // allocated size of any and all recorded oops
   csize_t total_oop_size() const {
     OopRecorder* recorder = oop_recorder();
-    return (recorder == NULL)? 0: recorder->oop_size();
+    return (recorder == nullptr)? 0: recorder->oop_size();
   }
 
   // allocated size of any and all recorded metadata
   csize_t total_metadata_size() const {
     OopRecorder* recorder = oop_recorder();
-    return (recorder == NULL)? 0: recorder->metadata_size();
+    return (recorder == nullptr)? 0: recorder->metadata_size();
   }
 
   // Configuration functions, called immediately after the CB is constructed.
@@ -631,15 +682,18 @@ class CodeBuffer: public StackObj {
 
   address last_insn() const { return _last_insn; }
   void set_last_insn(address a) { _last_insn = a; }
-  void clear_last_insn() { set_last_insn(NULL); }
+  void clear_last_insn() { set_last_insn(nullptr); }
+
+  address last_label() const { return _last_label; }
+  void set_last_label(address a) { _last_label = a; }
 
 #ifndef PRODUCT
-  CodeStrings& strings() { return _code_strings; }
+  AsmRemarks &asm_remarks() { return _asm_remarks; }
+  DbgStrings &dbg_strings() { return _dbg_strings; }
 
-  void free_strings() {
-    if (!_code_strings.is_null()) {
-      _code_strings.free(); // sets _strings Null as a side-effect.
-    }
+  void clear_strings() {
+    _asm_remarks.clear();
+    _dbg_strings.clear();
   }
 #endif
 
@@ -656,7 +710,7 @@ class CodeBuffer: public StackObj {
 
   // NMethod generation
   void copy_code_and_locs_to(CodeBlob* blob) {
-    assert(blob != NULL, "sane");
+    assert(blob != nullptr, "sane");
     copy_relocations_to(blob);
     copy_code_to(blob);
   }
@@ -666,12 +720,22 @@ class CodeBuffer: public StackObj {
     }
   }
 
-  void block_comment(intptr_t offset, const char * comment) PRODUCT_RETURN;
-  const char* code_string(const char* str) PRODUCT_RETURN_(return NULL;);
+  void block_comment(ptrdiff_t offset, const char* comment) PRODUCT_RETURN;
+  const char* code_string(const char* str) PRODUCT_RETURN_(return nullptr;);
 
   // Log a little info about section usage in the CodeBuffer
   void log_section_sizes(const char* name);
 
+  // Make a set of stubs final. It can create/optimize stubs.
+  bool finalize_stubs();
+
+  // Request for a shared stub to the interpreter
+  void shared_stub_to_interp_for(ciMethod* callee, csize_t call_offset);
+
+  void set_const_section_alignment(int align) {
+    _const_section_alignment = align_up(align, HeapWordSize);
+  }
+
 #ifndef PRODUCT
  public:
   // Printing / Decoding
@@ -687,6 +751,23 @@ class CodeBuffer: public StackObj {
 
 };
 
+// A Java method can have calls of Java methods which can be statically bound.
+// Calls of Java methods need stubs to the interpreter. Calls sharing the same Java method
+// can share a stub to the interpreter.
+// A SharedStubToInterpRequest is a request for a shared stub to the interpreter.
+class SharedStubToInterpRequest : public ResourceObj {
+ private:
+  ciMethod* _shared_method;
+  CodeBuffer::csize_t _call_offset; // The offset of the call in CodeBuffer
+
+ public:
+  SharedStubToInterpRequest(ciMethod* method = nullptr, CodeBuffer::csize_t call_offset = -1) : _shared_method(method),
+      _call_offset(call_offset) {}
+
+  ciMethod* shared_method() const { return _shared_method; }
+  CodeBuffer::csize_t call_offset() const { return _call_offset; }
+};
+
 inline bool CodeSection::maybe_expand_to_ensure_remaining(csize_t amount) {
   if (remaining() < amount) { _outer->expand(this, amount); return true; }
   return false;
