@@ -49,7 +49,7 @@ class Mutex;
 //               |        |    |
 //               |  data  |    |
 //               |        |    |
-// code_begin -->|--------|    |       <--- aligned by CodeEntryAlignment
+// code_begin -->|--------|    |       <--- aligned by code_alignment()
 //               |        |    |
 //               |        |    |
 //               |  code  |    | size
@@ -99,6 +99,7 @@ class StubInterface: public CHeapObj<mtCode> {
   // General info/converters
   virtual int     size(Stub* self) const                   = 0; // the total size of the stub in bytes (must be a multiple of HeapWordSize)
   virtual int     alignment() const                        = 0; // computes the alignment
+  virtual int     code_alignment() const                   = 0; // computes the code alignment
 
   // Code info
   virtual address code_begin(Stub* self) const             = 0; // points to the first code byte
@@ -127,6 +128,7 @@ class StubInterface: public CHeapObj<mtCode> {
     /* General info */                                     \
     virtual int     size(Stub* self) const                 { return cast(self)->size(); }          \
     virtual int     alignment() const                      { return stub::alignment(); }           \
+    virtual int     code_alignment() const                 { return stub::code_alignment(); }      \
                                                            \
     /* Code info */                                        \
     virtual address code_begin(Stub* self) const           { return cast(self)->code_begin(); }    \
@@ -154,9 +156,10 @@ class StubQueue: public CHeapObj<mtCode> {
   Mutex* const   _mutex;                         // the lock used for a (request, commit) transaction
 
   void  check_index(int i) const                 { assert(0 <= i && i < _buffer_limit && i % stub_alignment() == 0, "illegal index"); }
+  void  check_stub_align(Stub* s) const          { assert(((intptr_t)s) % stub_alignment() == 0, "incorrect stub alignment"); }
   bool  is_contiguous() const                    { return _queue_begin <= _queue_end; }
-  int   index_of(Stub* s) const                  { int i = (address)s - _stub_buffer; check_index(i); return i; }
-  Stub* stub_at(int i) const                     { check_index(i); return (Stub*)(_stub_buffer + i); }
+  int   index_of(Stub* s) const                  { int i = (int)((address)s - _stub_buffer); check_index(i); return i; }
+  Stub* stub_at(int i) const                     { check_index(i); Stub* s = (Stub*)(_stub_buffer + i); check_stub_align(s); return s; }
   Stub* current_stub() const                     { return stub_at(_queue_end); }
 
   // Stub functionality accessed via interface
