@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012, 2023, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2012, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -109,8 +109,6 @@ static void close_emergency_dump_file() {
 }
 
 static const char* create_emergency_dump_path() {
-  assert(is_path_empty(), "invariant");
-
   const size_t path_len = get_dump_directory();
   if (path_len == 0) {
     return nullptr;
@@ -497,7 +495,7 @@ static bool prepare_for_emergency_dump(Thread* thread) {
     Service_lock->unlock();
   }
 
-  if (UseNotificationThread && Notification_lock->owned_by_self()) {
+  if (Notification_lock->owned_by_self()) {
     Notification_lock->unlock();
   }
 
