@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 2023, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1997, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -27,6 +27,7 @@
 #include "gc/shared/c2/barrierSetC2.hpp"
 #include "memory/allocation.inline.hpp"
 #include "memory/resourceArea.hpp"
+#include "opto/addnode.hpp"
 #include "opto/block.hpp"
 #include "opto/callnode.hpp"
 #include "opto/castnode.hpp"
@@ -364,6 +365,7 @@ NodeHash::~NodeHash() {
 //------------------------------PhaseRemoveUseless-----------------------------
 // 1) Use a breadthfirst walk to collect useful nodes reachable from root.
 PhaseRemoveUseless::PhaseRemoveUseless(PhaseGVN* gvn, Unique_Node_List& worklist, PhaseNumber phase_num) : Phase(phase_num) {
+  C->print_method(PHASE_BEFORE_REMOVEUSELESS, 3);
   // Implementation requires an edge from root to each SafePointNode
   // at a backward branch. Inserted in add_safepoint().
 
@@ -669,17 +671,10 @@ Node* PhaseGVN::apply_ideal(Node* k, bool can_reshape) {
   return i;
 }
 
-//------------------------------transform--------------------------------------
-// Return a node which computes the same function as this node, but in a
-// faster or cheaper fashion.
-Node *PhaseGVN::transform( Node *n ) {
-  return transform_no_reclaim(n);
-}
-
 //------------------------------transform--------------------------------------
 // Return a node which computes the same function as this node, but
 // in a faster or cheaper fashion.
-Node *PhaseGVN::transform_no_reclaim(Node *n) {
+Node* PhaseGVN::transform(Node* n) {
   NOT_PRODUCT( set_transforms(); )
 
   // Apply the Ideal call in a loop until it no longer applies
@@ -691,7 +686,7 @@ Node *PhaseGVN::transform_no_reclaim(Node *n) {
     k = i;
 #ifdef ASSERT
     if (loop_count >= K + C->live_nodes()) {
-      dump_infinite_loop_info(i, "PhaseGVN::transform_no_reclaim");
+      dump_infinite_loop_info(i, "PhaseGVN::transform");
     }
 #endif
     i = apply_ideal(k, /*can_reshape=*/false);
@@ -786,7 +781,7 @@ void PhaseGVN::dead_loop_check( Node *n ) {
         }
       }
     }
-    if (!no_dead_loop) n->dump_bfs(100,0,"#");
+    if (!no_dead_loop) n->dump_bfs(100,nullptr,"#");
     assert(no_dead_loop, "dead loop detected");
   }
 }
@@ -891,9 +886,12 @@ void PhaseIterGVN::verify_step(Node* n) {
 }
 
 void PhaseIterGVN::trace_PhaseIterGVN(Node* n, Node* nn, const Type* oldtype) {
+  const Type* newtype = type_or_null(n);
+  if (nn != n || oldtype != newtype) {
+    C->print_method(PHASE_AFTER_ITER_GVN_STEP, 5, n);
+  }
   if (TraceIterativeGVN) {
     uint wlsize = _worklist.size();
-    const Type* newtype = type_or_null(n);
     if (nn != n) {
       // print old node
       tty->print("< ");
@@ -1021,6 +1019,8 @@ void PhaseIterGVN::trace_PhaseIterGVN_verbose(Node* n, int num_processed) {
 void PhaseIterGVN::optimize() {
   DEBUG_ONLY(uint num_processed  = 0;)
   NOT_PRODUCT(init_verifyPhaseIterGVN();)
+  NOT_PRODUCT(C->reset_igv_phase_iter(PHASE_AFTER_ITER_GVN_STEP);)
+  C->print_method(PHASE_BEFORE_ITER_GVN, 3);
   if (StressIGVN) {
     shuffle_worklist();
   }
@@ -1030,12 +1030,14 @@ void PhaseIterGVN::optimize() {
   // update edge info and put uses on worklist.
   while(_worklist.size()) {
     if (C->check_node_count(NodeLimitFudgeFactor * 2, "Out of nodes")) {
+      C->print_method(PHASE_AFTER_ITER_GVN, 3);
       return;
     }
     Node* n  = _worklist.pop();
     if (loop_count >= K * C->live_nodes()) {
       DEBUG_ONLY(dump_infinite_loop_info(n, "PhaseIterGVN::optimize");)
       C->record_method_not_compilable("infinite loop in PhaseIterGVN::optimize");
+      C->print_method(PHASE_AFTER_ITER_GVN, 3);
       return;
     }
     DEBUG_ONLY(trace_PhaseIterGVN_verbose(n, num_processed++);)
@@ -1050,6 +1052,7 @@ void PhaseIterGVN::optimize() {
     loop_count++;
   }
   NOT_PRODUCT(verify_PhaseIterGVN();)
+  C->print_method(PHASE_AFTER_ITER_GVN, 3);
 }
 
 #ifdef ASSERT
@@ -1133,7 +1136,7 @@ bool PhaseIterGVN::verify_node_value(Node* n) {
   }
   tty->cr();
   tty->print_cr("Missed Value optimization:");
-  n->dump_bfs(1, 0, "");
+  n->dump_bfs(1, nullptr, "");
   tty->print_cr("Current type:");
   told->dump_on(tty);
   tty->cr();
@@ -1441,9 +1444,9 @@ void PhaseIterGVN::subsume_node( Node *old, Node *nn ) {
 }
 
 //------------------------------add_users_to_worklist--------------------------
-void PhaseIterGVN::add_users_to_worklist0( Node *n ) {
+void PhaseIterGVN::add_users_to_worklist0(Node* n, Unique_Node_List& worklist) {
   for (DUIterator_Fast imax, i = n->fast_outs(imax); i < imax; i++) {
-    _worklist.push(n->fast_out(i));  // Push on worklist
+    worklist.push(n->fast_out(i));  // Push on worklist
   }
 }
 
@@ -1468,117 +1471,121 @@ static PhiNode* countedloop_phi_from_cmp(CmpNode* cmp, Node* n) {
   return nullptr;
 }
 
-void PhaseIterGVN::add_users_to_worklist( Node *n ) {
-  add_users_to_worklist0(n);
+void PhaseIterGVN::add_users_to_worklist(Node *n) {
+  add_users_to_worklist0(n, _worklist);
 
+  Unique_Node_List& worklist = _worklist;
   // Move users of node to worklist
   for (DUIterator_Fast imax, i = n->fast_outs(imax); i < imax; i++) {
     Node* use = n->fast_out(i); // Get use
+    add_users_of_use_to_worklist(n, use, worklist);
+  }
+}
 
-    if( use->is_Multi() ||      // Multi-definer?  Push projs on worklist
-        use->is_Store() )       // Enable store/load same address
-      add_users_to_worklist0(use);
+void PhaseIterGVN::add_users_of_use_to_worklist(Node* n, Node* use, Unique_Node_List& worklist) {
+  if(use->is_Multi() ||      // Multi-definer?  Push projs on worklist
+      use->is_Store() )       // Enable store/load same address
+    add_users_to_worklist0(use, worklist);
 
-    // If we changed the receiver type to a call, we need to revisit
-    // the Catch following the call.  It's looking for a non-null
-    // receiver to know when to enable the regular fall-through path
-    // in addition to the NullPtrException path.
-    if (use->is_CallDynamicJava() && n == use->in(TypeFunc::Parms)) {
-      Node* p = use->as_CallDynamicJava()->proj_out_or_null(TypeFunc::Control);
-      if (p != nullptr) {
-        add_users_to_worklist0(p);
-      }
+  // If we changed the receiver type to a call, we need to revisit
+  // the Catch following the call.  It's looking for a non-null
+  // receiver to know when to enable the regular fall-through path
+  // in addition to the NullPtrException path.
+  if (use->is_CallDynamicJava() && n == use->in(TypeFunc::Parms)) {
+    Node* p = use->as_CallDynamicJava()->proj_out_or_null(TypeFunc::Control);
+    if (p != nullptr) {
+      add_users_to_worklist0(p, worklist);
     }
+  }
 
-    uint use_op = use->Opcode();
-    if(use->is_Cmp()) {       // Enable CMP/BOOL optimization
-      add_users_to_worklist(use); // Put Bool on worklist
-      if (use->outcnt() > 0) {
-        Node* bol = use->raw_out(0);
-        if (bol->outcnt() > 0) {
-          Node* iff = bol->raw_out(0);
-          if (iff->outcnt() == 2) {
-            // Look for the 'is_x2logic' pattern: "x ? : 0 : 1" and put the
-            // phi merging either 0 or 1 onto the worklist
-            Node* ifproj0 = iff->raw_out(0);
-            Node* ifproj1 = iff->raw_out(1);
-            if (ifproj0->outcnt() > 0 && ifproj1->outcnt() > 0) {
-              Node* region0 = ifproj0->raw_out(0);
-              Node* region1 = ifproj1->raw_out(0);
-              if( region0 == region1 )
-                add_users_to_worklist0(region0);
-            }
+  uint use_op = use->Opcode();
+  if(use->is_Cmp()) {       // Enable CMP/BOOL optimization
+    add_users_to_worklist0(use, worklist); // Put Bool on worklist
+    if (use->outcnt() > 0) {
+      Node* bol = use->raw_out(0);
+      if (bol->outcnt() > 0) {
+        Node* iff = bol->raw_out(0);
+        if (iff->outcnt() == 2) {
+          // Look for the 'is_x2logic' pattern: "x ? : 0 : 1" and put the
+          // phi merging either 0 or 1 onto the worklist
+          Node* ifproj0 = iff->raw_out(0);
+          Node* ifproj1 = iff->raw_out(1);
+          if (ifproj0->outcnt() > 0 && ifproj1->outcnt() > 0) {
+            Node* region0 = ifproj0->raw_out(0);
+            Node* region1 = ifproj1->raw_out(0);
+            if( region0 == region1 )
+              add_users_to_worklist0(region0, worklist);
           }
         }
       }
-      if (use_op == Op_CmpI || use_op == Op_CmpL) {
-        Node* phi = countedloop_phi_from_cmp(use->as_Cmp(), n);
-        if (phi != nullptr) {
-          // Input to the cmp of a loop exit check has changed, thus
-          // the loop limit may have changed, which can then change the
-          // range values of the trip-count Phi.
-          _worklist.push(phi);
-        }
+    }
+    if (use_op == Op_CmpI || use_op == Op_CmpL) {
+      Node* phi = countedloop_phi_from_cmp(use->as_Cmp(), n);
+      if (phi != nullptr) {
+        // Input to the cmp of a loop exit check has changed, thus
+        // the loop limit may have changed, which can then change the
+        // range values of the trip-count Phi.
+        worklist.push(phi);
       }
-      if (use_op == Op_CmpI) {
-        Node* cmp = use;
-        Node* in1 = cmp->in(1);
-        Node* in2 = cmp->in(2);
-        // Notify CmpI / If pattern from CastIINode::Value (left pattern).
-        // Must also notify if in1 is modified and possibly turns into X (right pattern).
-        //
-        // in1  in2                   in1  in2
-        //  |    |                     |    |
-        //  +--- | --+                 |    |
-        //  |    |   |                 |    |
-        // CmpINode  |                CmpINode
-        //    |      |                   |
-        // BoolNode  |                BoolNode
-        //    |      |        OR         |
-        //  IfNode   |                 IfNode
-        //    |      |                   |
-        //  IfProj   |                 IfProj   X
-        //    |      |                   |      |
-        //   CastIINode                 CastIINode
-        //
-        if (in1 != in2) { // if they are equal, the CmpI can fold them away
-          if (in1 == n) {
-            // in1 modified -> could turn into X -> do traversal based on right pattern.
-            for (DUIterator_Fast i2max, i2 = cmp->fast_outs(i2max); i2 < i2max; i2++) {
-              Node* bol = cmp->fast_out(i2); // For each Bool
-              if (bol->is_Bool()) {
-                for (DUIterator_Fast i3max, i3 = bol->fast_outs(i3max); i3 < i3max; i3++) {
-                  Node* iff = bol->fast_out(i3); // For each If
-                  if (iff->is_If()) {
-                    for (DUIterator_Fast i4max, i4 = iff->fast_outs(i4max); i4 < i4max; i4++) {
-                      Node* if_proj = iff->fast_out(i4); // For each IfProj
-                      assert(if_proj->is_IfProj(), "If only has IfTrue and IfFalse as outputs");
-                      for (DUIterator_Fast i5max, i5 = if_proj->fast_outs(i5max); i5 < i5max; i5++) {
-                        Node* castii = if_proj->fast_out(i5); // For each CastII
-                        if (castii->is_CastII() &&
-                            castii->as_CastII()->carry_dependency()) {
-                          _worklist.push(castii);
-                        }
+    }
+    if (use_op == Op_CmpI) {
+      Node* cmp = use;
+      Node* in1 = cmp->in(1);
+      Node* in2 = cmp->in(2);
+      // Notify CmpI / If pattern from CastIINode::Value (left pattern).
+      // Must also notify if in1 is modified and possibly turns into X (right pattern).
+      //
+      // in1  in2                   in1  in2
+      //  |    |                     |    |
+      //  +--- | --+                 |    |
+      //  |    |   |                 |    |
+      // CmpINode  |                CmpINode
+      //    |      |                   |
+      // BoolNode  |                BoolNode
+      //    |      |        OR         |
+      //  IfNode   |                 IfNode
+      //    |      |                   |
+      //  IfProj   |                 IfProj   X
+      //    |      |                   |      |
+      //   CastIINode                 CastIINode
+      //
+      if (in1 != in2) { // if they are equal, the CmpI can fold them away
+        if (in1 == n) {
+          // in1 modified -> could turn into X -> do traversal based on right pattern.
+          for (DUIterator_Fast i2max, i2 = cmp->fast_outs(i2max); i2 < i2max; i2++) {
+            Node* bol = cmp->fast_out(i2); // For each Bool
+            if (bol->is_Bool()) {
+              for (DUIterator_Fast i3max, i3 = bol->fast_outs(i3max); i3 < i3max; i3++) {
+                Node* iff = bol->fast_out(i3); // For each If
+                if (iff->is_If()) {
+                  for (DUIterator_Fast i4max, i4 = iff->fast_outs(i4max); i4 < i4max; i4++) {
+                    Node* if_proj = iff->fast_out(i4); // For each IfProj
+                    assert(if_proj->is_IfProj(), "If only has IfTrue and IfFalse as outputs");
+                    for (DUIterator_Fast i5max, i5 = if_proj->fast_outs(i5max); i5 < i5max; i5++) {
+                      Node* castii = if_proj->fast_out(i5); // For each CastII
+                      if (castii->is_CastII() &&
+                          castii->as_CastII()->carry_dependency()) {
+                        worklist.push(castii);
                       }
                     }
                   }
                 }
               }
             }
-          } else {
-            // Only in2 modified -> can assume X == in2 (left pattern).
-            assert(n == in2, "only in2 modified");
-            // Find all CastII with input in1.
-            for (DUIterator_Fast jmax, j = in1->fast_outs(jmax); j < jmax; j++) {
-              Node* castii = in1->fast_out(j);
-              if (castii->is_CastII() && castii->as_CastII()->carry_dependency()) {
-                // Find If.
-                if (castii->in(0) != nullptr && castii->in(0)->in(0) != nullptr && castii->in(0)->in(0)->is_If()) {
-                  Node* ifnode = castii->in(0)->in(0);
-                  // Check that if connects to the cmp
-                  if (ifnode->in(1) != nullptr && ifnode->in(1)->is_Bool() && ifnode->in(1)->in(1) == cmp) {
-                    _worklist.push(castii);
-                  }
+          }
+        } else {
+          // Only in2 modified -> can assume X == in2 (left pattern).
+          assert(n == in2, "only in2 modified");
+          // Find all CastII with input in1.
+          for (DUIterator_Fast jmax, j = in1->fast_outs(jmax); j < jmax; j++) {
+            Node* castii = in1->fast_out(j);
+            if (castii->is_CastII() && castii->as_CastII()->carry_dependency()) {
+              // Find If.
+              if (castii->in(0) != nullptr && castii->in(0)->in(0) != nullptr && castii->in(0)->in(0)->is_If()) {
+                Node* ifnode = castii->in(0)->in(0);
+                // Check that if connects to the cmp
+                if (ifnode->in(1) != nullptr && ifnode->in(1)->is_Bool() && ifnode->in(1)->in(1) == cmp) {
+                  worklist.push(castii);
                 }
               }
             }
@@ -1586,104 +1593,121 @@ void PhaseIterGVN::add_users_to_worklist( Node *n ) {
         }
       }
     }
+  }
 
-    // If changed Cast input, notify down for Phi and Sub - both do "uncast"
-    // Patterns:
-    // ConstraintCast+ -> Sub
-    // ConstraintCast+ -> Phi
-    if (use->is_ConstraintCast()) {
-      auto push_phi_or_sub_uses_to_worklist = [&](Node* n){
-        if (n->is_Phi() || n->is_Sub()) {
-          _worklist.push(n);
-        }
-      };
-      ConstraintCastNode::visit_uncasted_uses(use, push_phi_or_sub_uses_to_worklist);
+  // If changed Cast input, notify down for Phi, Sub, and Xor - all do "uncast"
+  // Patterns:
+  // ConstraintCast+ -> Sub
+  // ConstraintCast+ -> Phi
+  // ConstraintCast+ -> Xor
+  if (use->is_ConstraintCast()) {
+    auto push_the_uses_to_worklist = [&](Node* n){
+      if (n->is_Phi() || n->is_Sub() || n->Opcode() == Op_XorI || n->Opcode() == Op_XorL) {
+        worklist.push(n);
+      }
+    };
+    auto is_boundary = [](Node* n){ return !n->is_ConstraintCast(); };
+    use->visit_uses(push_the_uses_to_worklist, is_boundary);
+  }
+  // If changed LShift inputs, check RShift users for useless sign-ext
+  if( use_op == Op_LShiftI ) {
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      if (u->Opcode() == Op_RShiftI)
+        worklist.push(u);
     }
-    // If changed LShift inputs, check RShift users for useless sign-ext
-    if( use_op == Op_LShiftI ) {
-      for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
-        Node* u = use->fast_out(i2);
-        if (u->Opcode() == Op_RShiftI)
-          _worklist.push(u);
+  }
+  // If changed LShift inputs, check And users for shift and mask (And) operation
+  if (use_op == Op_LShiftI || use_op == Op_LShiftL) {
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      if (u->Opcode() == Op_AndI || u->Opcode() == Op_AndL) {
+        worklist.push(u);
       }
     }
-    // If changed LShift inputs, check And users for shift and mask (And) operation
-    if (use_op == Op_LShiftI || use_op == Op_LShiftL) {
-      for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
-        Node* u = use->fast_out(i2);
-        if (u->Opcode() == Op_AndI || u->Opcode() == Op_AndL) {
-          _worklist.push(u);
-        }
+  }
+  // If changed AddI/SubI inputs, check CmpU for range check optimization.
+  if (use_op == Op_AddI || use_op == Op_SubI) {
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      if (u->is_Cmp() && (u->Opcode() == Op_CmpU)) {
+        worklist.push(u);
       }
     }
-    // If changed AddI/SubI inputs, check CmpU for range check optimization.
-    if (use_op == Op_AddI || use_op == Op_SubI) {
-      for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
-        Node* u = use->fast_out(i2);
-        if (u->is_Cmp() && (u->Opcode() == Op_CmpU)) {
-          _worklist.push(u);
-        }
+  }
+  // If changed AddP inputs:
+  // - check Stores for loop invariant, and
+  // - if the changed input is the offset, check constant-offset AddP users for
+  //   address expression flattening.
+  if (use_op == Op_AddP) {
+    bool offset_changed = n == use->in(AddPNode::Offset);
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      if (u->is_Mem()) {
+        worklist.push(u);
+      } else if (offset_changed && u->is_AddP() && u->in(AddPNode::Offset)->is_Con()) {
+        worklist.push(u);
       }
     }
-    // If changed AddP inputs, check Stores for loop invariant
-    if( use_op == Op_AddP ) {
-      for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
-        Node* u = use->fast_out(i2);
-        if (u->is_Mem())
-          _worklist.push(u);
-      }
+  }
+  // If changed initialization activity, check dependent Stores
+  if (use_op == Op_Allocate || use_op == Op_AllocateArray) {
+    InitializeNode* init = use->as_Allocate()->initialization();
+    if (init != nullptr) {
+      Node* imem = init->proj_out_or_null(TypeFunc::Memory);
+      if (imem != nullptr) add_users_to_worklist0(imem, worklist);
     }
-    // If changed initialization activity, check dependent Stores
-    if (use_op == Op_Allocate || use_op == Op_AllocateArray) {
-      InitializeNode* init = use->as_Allocate()->initialization();
-      if (init != nullptr) {
-        Node* imem = init->proj_out_or_null(TypeFunc::Memory);
-        if (imem != nullptr)  add_users_to_worklist0(imem);
-      }
-    }
-    // If the ValidLengthTest input changes then the fallthrough path out of the AllocateArray may have become dead.
-    // CatchNode::Value() is responsible for killing that path. The CatchNode has to be explicitly enqueued for igvn
-    // to guarantee the change is not missed.
-    if (use_op == Op_AllocateArray && n == use->in(AllocateNode::ValidLengthTest)) {
-      Node* p = use->as_AllocateArray()->proj_out_or_null(TypeFunc::Control);
-      if (p != nullptr) {
-        add_users_to_worklist0(p);
-      }
+  }
+  // If the ValidLengthTest input changes then the fallthrough path out of the AllocateArray may have become dead.
+  // CatchNode::Value() is responsible for killing that path. The CatchNode has to be explicitly enqueued for igvn
+  // to guarantee the change is not missed.
+  if (use_op == Op_AllocateArray && n == use->in(AllocateNode::ValidLengthTest)) {
+    Node* p = use->as_AllocateArray()->proj_out_or_null(TypeFunc::Control);
+    if (p != nullptr) {
+      add_users_to_worklist0(p, worklist);
     }
+  }
 
-    if (use_op == Op_Initialize) {
-      Node* imem = use->as_Initialize()->proj_out_or_null(TypeFunc::Memory);
-      if (imem != nullptr)  add_users_to_worklist0(imem);
-    }
-    // Loading the java mirror from a Klass requires two loads and the type
-    // of the mirror load depends on the type of 'n'. See LoadNode::Value().
-    //   LoadBarrier?(LoadP(LoadP(AddP(foo:Klass, #java_mirror))))
-    BarrierSetC2* bs = BarrierSet::barrier_set()->barrier_set_c2();
-    bool has_load_barrier_nodes = bs->has_load_barrier_nodes();
+  if (use_op == Op_Initialize) {
+    Node* imem = use->as_Initialize()->proj_out_or_null(TypeFunc::Memory);
+    if (imem != nullptr) add_users_to_worklist0(imem, worklist);
+  }
+  // Loading the java mirror from a Klass requires two loads and the type
+  // of the mirror load depends on the type of 'n'. See LoadNode::Value().
+  //   LoadBarrier?(LoadP(LoadP(AddP(foo:Klass, #java_mirror))))
+  BarrierSetC2* bs = BarrierSet::barrier_set()->barrier_set_c2();
+  bool has_load_barrier_nodes = bs->has_load_barrier_nodes();
 
-    if (use_op == Op_LoadP && use->bottom_type()->isa_rawptr()) {
-      for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
-        Node* u = use->fast_out(i2);
-        const Type* ut = u->bottom_type();
-        if (u->Opcode() == Op_LoadP && ut->isa_instptr()) {
-          if (has_load_barrier_nodes) {
-            // Search for load barriers behind the load
-            for (DUIterator_Fast i3max, i3 = u->fast_outs(i3max); i3 < i3max; i3++) {
-              Node* b = u->fast_out(i3);
-              if (bs->is_gc_barrier_node(b)) {
-                _worklist.push(b);
-              }
+  if (use_op == Op_LoadP && use->bottom_type()->isa_rawptr()) {
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      const Type* ut = u->bottom_type();
+      if (u->Opcode() == Op_LoadP && ut->isa_instptr()) {
+        if (has_load_barrier_nodes) {
+          // Search for load barriers behind the load
+          for (DUIterator_Fast i3max, i3 = u->fast_outs(i3max); i3 < i3max; i3++) {
+            Node* b = u->fast_out(i3);
+            if (bs->is_gc_barrier_node(b)) {
+              worklist.push(b);
             }
           }
-          _worklist.push(u);
         }
+        worklist.push(u);
       }
     }
-    if (use->Opcode() == Op_OpaqueZeroTripGuard) {
-      assert(use->outcnt() <= 1, "OpaqueZeroTripGuard can't be shared");
-      if (use->outcnt() == 1) {
-        Node* cmp = use->unique_out();
-        _worklist.push(cmp);
+  }
+  if (use->Opcode() == Op_OpaqueZeroTripGuard) {
+    assert(use->outcnt() <= 1, "OpaqueZeroTripGuard can't be shared");
+    if (use->outcnt() == 1) {
+      Node* cmp = use->unique_out();
+      worklist.push(cmp);
+    }
+  }
+  if (use->Opcode() == Op_AddX) {
+    for (DUIterator_Fast i2max, i2 = use->fast_outs(i2max); i2 < i2max; i2++) {
+      Node* u = use->fast_out(i2);
+      if (u->Opcode() == Op_CastX2P) {
+        worklist.push(u);
       }
     }
   }
@@ -1823,7 +1847,7 @@ void PhaseCCP::verify_analyze(Unique_Node_List& worklist_verify) {
   // We should either make sure that these nodes are properly added back to the CCP worklist
   // in PhaseCCP::push_child_nodes_to_worklist() to update their type or add an exception
   // in the verification code above if that is not possible for some reason (like Load nodes).
-  assert(!failure, "Missed optimization opportunity in PhaseCCP");
+  assert(!failure, "PhaseCCP not at fixpoint: analysis result may be unsound.");
 }
 #endif
 
@@ -1876,6 +1900,7 @@ void PhaseCCP::push_more_uses(Unique_Node_List& worklist, Node* parent, const No
   push_and(worklist, parent, use);
   push_cast_ii(worklist, parent, use);
   push_opaque_zero_trip_guard(worklist, use);
+  push_bool_with_cmpu_and_mask(worklist, use);
 }
 
 
@@ -1922,6 +1947,57 @@ void PhaseCCP::push_cmpu(Unique_Node_List& worklist, const Node* use) const {
   }
 }
 
+// Look for the following shape, which can be optimized by BoolNode::Value_cmpu_and_mask() (i.e. corresponds to case
+// (1b): "(m & x) <u (m + 1))".
+// If any of the inputs on the level (%%) change, we need to revisit Bool because we could have prematurely found that
+// the Bool is constant (i.e. case (1b) can be applied) which could become invalid with new type information during CCP.
+//
+//  m    x  m    1  (%%)
+//   \  /    \  /
+//   AndI    AddI
+//      \    /
+//       CmpU
+//        |
+//       Bool
+//
+void PhaseCCP::push_bool_with_cmpu_and_mask(Unique_Node_List& worklist, const Node* use) const {
+  uint use_op = use->Opcode();
+  if (use_op != Op_AndI && (use_op != Op_AddI || use->in(2)->find_int_con(0) != 1)) {
+    // Not "m & x" or "m + 1"
+    return;
+  }
+  for (DUIterator_Fast imax, i = use->fast_outs(imax); i < imax; i++) {
+    Node* cmpu = use->fast_out(i);
+    if (cmpu->Opcode() == Op_CmpU) {
+      push_bool_matching_case1b(worklist, cmpu);
+    }
+  }
+}
+
+// Push any Bool below 'cmpu' that matches case (1b) of BoolNode::Value_cmpu_and_mask().
+void PhaseCCP::push_bool_matching_case1b(Unique_Node_List& worklist, const Node* cmpu) const {
+  assert(cmpu->Opcode() == Op_CmpU, "must be");
+  for (DUIterator_Fast imax, i = cmpu->fast_outs(imax); i < imax; i++) {
+    Node* bol = cmpu->fast_out(i);
+    if (!bol->is_Bool() || bol->as_Bool()->_test._test != BoolTest::lt) {
+      // Not a Bool with "<u"
+      continue;
+    }
+    Node* andI = cmpu->in(1);
+    Node* addI = cmpu->in(2);
+    if (andI->Opcode() != Op_AndI || addI->Opcode() != Op_AddI || addI->in(2)->find_int_con(0) != 1) {
+      // Not "m & x" and "m + 1"
+      continue;
+    }
+
+    Node* m = addI->in(1);
+    if (m == andI->in(1) || m == andI->in(2)) {
+      // Is "m" shared? Matched (1b) and thus we revisit Bool.
+      push_if_not_bottom_type(worklist, bol);
+    }
+  }
+}
+
 // If n is used in a counted loop exit condition, then the type of the counted loop's Phi depends on the type of 'n'.
 // Seem PhiNode::Value().
 void PhaseCCP::push_counted_loop_phi(Unique_Node_List& worklist, Node* parent, const Node* use) {
@@ -1966,7 +2042,7 @@ void PhaseCCP::push_load_barrier(Unique_Node_List& worklist, const BarrierSetC2*
 
 // AndI/L::Value() optimizes patterns similar to (v << 2) & 3 to zero if they are bitwise disjoint.
 // Add the AndI/L nodes back to the worklist to re-apply Value() in case the shift value changed.
-// Pattern: parent -> LShift (use) -> ConstraintCast* -> And
+// Pattern: parent -> LShift (use) -> (ConstraintCast | ConvI2L)* -> And
 void PhaseCCP::push_and(Unique_Node_List& worklist, const Node* parent, const Node* use) const {
   uint use_op = use->Opcode();
   if ((use_op == Op_LShiftI || use_op == Op_LShiftL)
@@ -1977,7 +2053,10 @@ void PhaseCCP::push_and(Unique_Node_List& worklist, const Node* parent, const No
         push_if_not_bottom_type(worklist, n);
       }
     };
-    ConstraintCastNode::visit_uncasted_uses(use, push_and_uses_to_worklist);
+    auto is_boundary = [](Node* n) {
+      return !(n->is_ConstraintCast() || n->Opcode() == Op_ConvI2L);
+    };
+    use->visit_uses(push_and_uses_to_worklist, is_boundary);
   }
 }
 
@@ -2263,7 +2342,15 @@ void PhasePeephole::print_statistics() {
 //------------------------------set_req_X--------------------------------------
 void Node::set_req_X( uint i, Node *n, PhaseIterGVN *igvn ) {
   assert( is_not_dead(n), "can not use dead node");
-  assert( igvn->hash_find(this) != this, "Need to remove from hash before changing edges" );
+#ifdef ASSERT
+  if (igvn->hash_find(this) == this) {
+    tty->print_cr("Need to remove from hash before changing edges");
+    this->dump(1);
+    tty->print_cr("Set at i = %d", i);
+    n->dump();
+    assert(false, "Need to remove from hash before changing edges");
+  }
+#endif
   Node *old = in(i);
   set_req(i, n);
 
@@ -2332,6 +2419,7 @@ void Node::replace_by(Node *new_node) {
 //=============================================================================
 //-----------------------------------------------------------------------------
 void Type_Array::grow( uint i ) {
+  assert(_a == Compile::current()->comp_arena(), "Should be allocated in comp_arena");
   if( !_max ) {
     _max = 1;
     _types = (const Type**)_a->Amalloc( _max * sizeof(Type*) );
