@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 2023, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 1997, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -643,6 +643,14 @@ CmpNode *CmpNode::make(Node *in1, Node *in2, BasicType bt, bool unsigned_comp) {
         return new CmpULNode(in1, in2);
       }
       return new CmpLNode(in1, in2);
+    case T_OBJECT:
+    case T_ARRAY:
+    case T_ADDRESS:
+    case T_METADATA:
+      return new CmpPNode(in1, in2);
+    case T_NARROWOOP:
+    case T_NARROWKLASS:
+      return new CmpNNode(in1, in2);
     default:
       fatal("Not implemented for %s", type2name(bt));
   }
@@ -998,8 +1006,8 @@ const Type *CmpPNode::sub( const Type *t1, const Type *t2 ) const {
     if (p0 && p1) {
       Node* in1 = in(1)->uncast();
       Node* in2 = in(2)->uncast();
-      AllocateNode* alloc1 = AllocateNode::Ideal_allocation(in1, nullptr);
-      AllocateNode* alloc2 = AllocateNode::Ideal_allocation(in2, nullptr);
+      AllocateNode* alloc1 = AllocateNode::Ideal_allocation(in1);
+      AllocateNode* alloc2 = AllocateNode::Ideal_allocation(in2);
       if (MemNode::detect_ptr_independence(in1, alloc1, in2, alloc2, nullptr)) {
         return TypeInt::CC_GT;  // different pointers
       }
@@ -1615,78 +1623,9 @@ Node *BoolNode::Ideal(PhaseGVN *phase, bool can_reshape) {
     return new BoolNode( ncmp, _test.negate() );
   }
 
-  // We use the following Lemmas/insights for the following two transformations (1) and (2):
-  //   x & y <=u y, for any x and y           (Lemma 1, masking always results in a smaller unsigned number)
-  //   y <u y + 1 is always true if y != -1   (Lemma 2, (uint)(-1 + 1) == (uint)(UINT_MAX + 1) which overflows)
-  //   y <u 0 is always false for any y       (Lemma 3, 0 == UINT_MIN and nothing can be smaller than that)
-  //
-  // (1a) Always:     Change ((x & m) <=u m  ) or ((m & x) <=u m  ) to always true   (true by Lemma 1)
-  // (1b) If m != -1: Change ((x & m) <u  m + 1) or ((m & x) <u  m + 1) to always true:
-  //    x & m <=u m          is always true   // (Lemma 1)
-  //    x & m <=u m <u m + 1 is always true   // (Lemma 2: m <u m + 1, if m != -1)
-  //
-  // A counter example for (1b), if we allowed m == -1:
-  //     (x & m)  <u m + 1
-  //     (x & -1) <u 0
-  //      x       <u 0
-  //   which is false for any x (Lemma 3)
-  if (cop == Op_CmpU &&
-      cmp1_op == Op_AndI) {
-    Node* m = nullptr;
-    if (_test._test == BoolTest::le) {
-      // (1a) "((x & m) <=u m)", cmp2 = m
-      m = cmp2;
-    } else if (_test._test == BoolTest::lt &&
-               cmp2->Opcode() == Op_AddI &&
-               cmp2->in(2)->find_int_con(0) == 1) {
-      // (1b) "(x & m) <u m + 1" and "(m & x) <u m + 1", cmp2 = m + 1
-      Node* rhs_m = cmp2->in(1);
-      const TypeInt* rhs_m_type = phase->type(rhs_m)->isa_int();
-      if (rhs_m_type->_lo > -1 || rhs_m_type->_hi < -1) {
-        // Exclude any case where m == -1 is possible.
-        m = rhs_m;
-      }
-    }
-    if (cmp1->in(2) == m || cmp1->in(1) == m) {
-      return ConINode::make(1);
-    }
-  }
-
-  // (2) Change ((x & (m - 1)) <u m) or (((m - 1) & x) <u m) to (m >u 0)
-  // This is the off-by-one variant of the above.
-  //
-  // We now prove that this replacement is correct. This is the same as proving
-  //   "m >u 0" if and only if "x & (m - 1) <u m", i.e. "m >u 0 <=> x & (m - 1) <u m"
-  //
-  // We use (Lemma 1) and (Lemma 3) from above.
-  //
-  // Case "x & (m - 1) <u m => m >u 0":
-  //   We prove this by contradiction:
-  //     Assume m <=u 0 which is equivalent to m == 0:
-  //   and thus
-  //     x & (m - 1) <u m = 0               // m == 0
-  //     y           <u     0               // y = x & (m - 1)
-  //   by Lemma 3, this is always false, i.e. a contradiction to our assumption.
-  //
-  // Case "m >u 0 => x & (m - 1) <u m":
-  //   x & (m - 1) <=u (m - 1)              // (Lemma 1)
-  //   x & (m - 1) <=u (m - 1) <u m         // Using assumption m >u 0, no underflow of "m - 1"
-  //
-  //
-  // Note that the signed version of "m > 0":
-  //   m > 0 <=> x & (m - 1) <u m
-  // does not hold:
-  //   Assume m == -1 and x == -1:
-  //     x  & (m - 1) <u m
-  //     -1 & -2      <u -1
-  //     -2           <u -1
-  //     UINT_MAX - 1 <u UINT_MAX           // Signed to unsigned numbers
-  // which is true while
-  //   m > 0
-  // is false which is a contradiction.
-  if (cop == Op_CmpU &&
-      _test._test == BoolTest::lt &&
-      cmp1_op == Op_AndI) {
+  // Transform: "((x & (m - 1)) <u m)" or "(((m - 1) & x) <u m)" into "(m >u 0)"
+  // This is case [CMPU_MASK] which is further described at the method comment of BoolNode::Value_cmpu_and_mask().
+  if (cop == Op_CmpU && _test._test == BoolTest::lt && cmp1_op == Op_AndI) {
     Node* m = cmp2; // RHS: m
     for (int add_idx = 1; add_idx <= 2; add_idx++) { // LHS: "(m + (-1)) & x" or "x & (m + (-1))"?
       Node* maybe_m_minus_1 = cmp1->in(add_idx);
@@ -1860,11 +1799,100 @@ Node *BoolNode::Ideal(PhaseGVN *phase, bool can_reshape) {
   //    }
 }
 
-//------------------------------Value------------------------------------------
+// We use the following Lemmas/insights for the following two transformations (1) and (2):
+//   x & y <=u y, for any x and y           (Lemma 1, masking always results in a smaller unsigned number)
+//   y <u y + 1 is always true if y != -1   (Lemma 2, (uint)(-1 + 1) == (uint)(UINT_MAX + 1) which overflows)
+//   y <u 0 is always false for any y       (Lemma 3, 0 == UINT_MIN and nothing can be smaller than that)
+//
+// (1a) Always:     Change ((x & m) <=u m  ) or ((m & x) <=u m  ) to always true   (true by Lemma 1)
+// (1b) If m != -1: Change ((x & m) <u  m + 1) or ((m & x) <u  m + 1) to always true:
+//    x & m <=u m          is always true   // (Lemma 1)
+//    x & m <=u m <u m + 1 is always true   // (Lemma 2: m <u m + 1, if m != -1)
+//
+// A counter example for (1b), if we allowed m == -1:
+//     (x & m)  <u m + 1
+//     (x & -1) <u 0
+//      x       <u 0
+//   which is false for any x (Lemma 3)
+//
+// (2) Change ((x & (m - 1)) <u m) or (((m - 1) & x) <u m) to (m >u 0)
+// This is the off-by-one variant of the above.
+//
+// We now prove that this replacement is correct. This is the same as proving
+//   "m >u 0" if and only if "x & (m - 1) <u m", i.e. "m >u 0 <=> x & (m - 1) <u m"
+//
+// We use (Lemma 1) and (Lemma 3) from above.
+//
+// Case "x & (m - 1) <u m => m >u 0":
+//   We prove this by contradiction:
+//     Assume m <=u 0 which is equivalent to m == 0:
+//   and thus
+//     x & (m - 1) <u m = 0               // m == 0
+//     y           <u     0               // y = x & (m - 1)
+//   by Lemma 3, this is always false, i.e. a contradiction to our assumption.
+//
+// Case "m >u 0 => x & (m - 1) <u m":
+//   x & (m - 1) <=u (m - 1)              // (Lemma 1)
+//   x & (m - 1) <=u (m - 1) <u m         // Using assumption m >u 0, no underflow of "m - 1"
+//
+//
+// Note that the signed version of "m > 0":
+//   m > 0 <=> x & (m - 1) <u m
+// does not hold:
+//   Assume m == -1 and x == -1:
+//     x  & (m - 1) <u m
+//     -1 & -2      <u -1
+//     -2           <u -1
+//     UINT_MAX - 1 <u UINT_MAX           // Signed to unsigned numbers
+// which is true while
+//   m > 0
+// is false which is a contradiction.
+//
+// (1a) and (1b) is covered by this method since we can directly return a true value as type while (2) is covered
+// in BoolNode::Ideal since we create a new non-constant node (see [CMPU_MASK]).
+const Type* BoolNode::Value_cmpu_and_mask(PhaseValues* phase) const {
+  Node* cmp = in(1);
+  if (cmp != nullptr && cmp->Opcode() == Op_CmpU) {
+    Node* cmp1 = cmp->in(1);
+    Node* cmp2 = cmp->in(2);
+
+    if (cmp1->Opcode() == Op_AndI) {
+      Node* m = nullptr;
+      if (_test._test == BoolTest::le) {
+        // (1a) "((x & m) <=u m)", cmp2 = m
+        m = cmp2;
+      } else if (_test._test == BoolTest::lt && cmp2->Opcode() == Op_AddI && cmp2->in(2)->find_int_con(0) == 1) {
+        // (1b) "(x & m) <u m + 1" and "(m & x) <u m + 1", cmp2 = m + 1
+        Node* rhs_m = cmp2->in(1);
+        const TypeInt* rhs_m_type = phase->type(rhs_m)->isa_int();
+        if (rhs_m_type != nullptr && (rhs_m_type->_lo > -1 || rhs_m_type->_hi < -1)) {
+          // Exclude any case where m == -1 is possible.
+          m = rhs_m;
+        }
+      }
+
+      if (cmp1->in(2) == m || cmp1->in(1) == m) {
+        return TypeInt::ONE;
+      }
+    }
+  }
+
+  return nullptr;
+}
+
 // Simplify a Bool (convert condition codes to boolean (1 or 0)) node,
 // based on local information.   If the input is constant, do it.
 const Type* BoolNode::Value(PhaseGVN* phase) const {
-  return _test.cc2logical( phase->type( in(1) ) );
+  const Type* input_type = phase->type(in(1));
+  if (input_type == Type::TOP) {
+    return Type::TOP;
+  }
+  const Type* t = Value_cmpu_and_mask(phase);
+  if (t != nullptr) {
+    return t;
+  }
+
+  return _test.cc2logical(input_type);
 }
 
 #ifndef PRODUCT
