@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019, 2022, Oracle and/or its affiliates. All rights reserved.
+ * Copyright (c) 2019, 2024, Oracle and/or its affiliates. All rights reserved.
  * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
  *
  * This code is free software; you can redistribute it and/or modify it
@@ -23,17 +23,23 @@
 package jdk.jpackage.test;
 
 import java.io.IOException;
+import java.lang.reflect.Method;
 import java.nio.file.Files;
 import java.nio.file.Path;
 import java.util.HashMap;
 import java.util.List;
 import java.util.Map;
+import java.util.Objects;
 import java.util.Optional;
 import java.util.Set;
 import java.util.function.BiConsumer;
+import java.util.regex.Matcher;
+import java.util.regex.Pattern;
 import java.util.stream.Collectors;
 import java.util.stream.Stream;
-import jdk.jpackage.test.Functional.ThrowingRunnable;
+import static jdk.jpackage.internal.util.function.ExceptionBox.rethrowUnchecked;
+import jdk.jpackage.internal.util.function.ThrowingRunnable;
+import static jdk.jpackage.internal.util.function.ThrowingSupplier.toSupplier;
 import jdk.jpackage.test.PackageTest.PackageHandlers;
 
 public class WindowsHelper {
@@ -91,8 +97,9 @@ public class WindowsHelper {
     static PackageHandlers createMsiPackageHandlers() {
         BiConsumer<JPackageCommand, Boolean> installMsi = (cmd, install) -> {
             cmd.verifyIsOfType(PackageType.WIN_MSI);
+            var msiPath = TransientMsi.create(cmd).path();
             runMsiexecWithRetries(Executor.of("msiexec", "/qn", "/norestart",
-                    install ? "/i" : "/x").addArgument(cmd.outputBundle().normalize()));
+                    install ? "/i" : "/x").addArgument(msiPath));
         };
 
         PackageHandlers msi = new PackageHandlers();
@@ -109,6 +116,8 @@ public class WindowsHelper {
                     TKit.removeRootFromAbsolutePath(
                             getInstallationRootDirectory(cmd)));
 
+            final Path msiPath = TransientMsi.create(cmd).path();
+
             // Put msiexec in .bat file because can't pass value of TARGETDIR
             // property containing spaces through ProcessBuilder properly.
             // Set folder permissions to allow msiexec unpack msi bundle.
@@ -118,16 +127,83 @@ public class WindowsHelper {
                     String.join(" ", List.of(
                     "msiexec",
                     "/a",
-                    String.format("\"%s\"", cmd.outputBundle().normalize()),
+                    String.format("\"%s\"", msiPath),
                     "/qn",
                     String.format("TARGETDIR=\"%s\"",
                             unpackDir.toAbsolutePath().normalize())))));
             runMsiexecWithRetries(Executor.of("cmd", "/c", unpackBat.toString()));
+
+            //
+            // WiX3 uses "." as the value of "DefaultDir" field for "ProgramFiles64Folder" folder in msi's Directory table
+            // WiX4 uses "PFiles64" as the value of "DefaultDir" field for "ProgramFiles64Folder" folder in msi's Directory table
+            // msiexec creates "Program Files/./<App Installation Directory>" from WiX3 msi which translates to "Program Files/<App Installation Directory>"
+            // msiexec creates "Program Files/PFiles64/<App Installation Directory>" from WiX4 msi
+            // So for WiX4 msi we need to transform "Program Files/PFiles64/<App Installation Directory>" into "Program Files/<App Installation Directory>"
+            //
+            // WiX4 does the same thing for %LocalAppData%.
+            //
+            for (var extraPathComponent : List.of("PFiles64", "LocalApp")) {
+                if (Files.isDirectory(unpackDir.resolve(extraPathComponent))) {
+                    Path installationSubDirectory = getInstallationSubDirectory(cmd);
+                    Path from = Path.of(extraPathComponent).resolve(installationSubDirectory);
+                    Path to = installationSubDirectory;
+                    TKit.trace(String.format("Convert [%s] into [%s] in [%s] directory", from, to,
+                            unpackDir));
+                    ThrowingRunnable.toRunnable(() -> {
+                        Files.createDirectories(unpackDir.resolve(to).getParent());
+                        Files.move(unpackDir.resolve(from), unpackDir.resolve(to));
+                        TKit.deleteDirectoryRecursive(unpackDir.resolve(extraPathComponent));
+                    }).run();
+                }
+            }
             return destinationDir;
         };
         return msi;
     }
 
+    record TransientMsi(Path path) {
+        static TransientMsi create(JPackageCommand cmd) {
+            var outputMsiPath = cmd.outputBundle().normalize();
+            if (isPathTooLong(outputMsiPath)) {
+                return toSupplier(() -> {
+                    var transientMsiPath = TKit.createTempDirectory("msi-copy").resolve("a.msi").normalize();
+                    TKit.trace(String.format("Copy [%s] to [%s]", outputMsiPath, transientMsiPath));
+                    Files.copy(outputMsiPath, transientMsiPath);
+                    return new TransientMsi(transientMsiPath);
+                }).get();
+            } else {
+                return new TransientMsi(outputMsiPath);
+            }
+        }
+    }
+
+    public enum WixType {
+        WIX3,
+        WIX4
+    }
+
+    public static WixType getWixTypeFromVerboseJPackageOutput(Executor.Result result) {
+        return result.getOutput().stream().map(str -> {
+            if (str.contains("[light.exe]")) {
+                return WixType.WIX3;
+            } else if (str.contains("[wix.exe]")) {
+                return WixType.WIX4;
+            } else {
+                return null;
+            }
+        }).filter(Objects::nonNull).reduce((a, b) -> {
+            throw new IllegalArgumentException("Invalid input: multiple invocations of WiX tools");
+        }).orElseThrow(() -> new IllegalArgumentException("Invalid input: no invocations of WiX tools"));
+    }
+
+    static Optional<Path> toShortPath(Path path) {
+        if (isPathTooLong(path)) {
+            return Optional.of(ShortPathUtils.toShortPath(path));
+        } else {
+            return Optional.empty();
+        }
+    }
+
     static PackageHandlers createExePackageHandlers() {
         BiConsumer<JPackageCommand, Boolean> installExe = (cmd, install) -> {
             cmd.verifyIsOfType(PackageType.WIN_EXE);
@@ -188,10 +264,97 @@ public class WindowsHelper {
                 "Failed to get file description of [%s]", pathToExeFile));
     }
 
+    public static void killProcess(long pid) {
+        Executor.of("taskkill", "/F", "/PID", Long.toString(pid)).dumpOutput(true).execute();
+    }
+
+    public static void killAppLauncherProcess(JPackageCommand cmd,
+            String launcherName, int expectedCount) {
+        var pids = findAppLauncherPIDs(cmd, launcherName);
+        try {
+            TKit.assertEquals(expectedCount, pids.length, String.format(
+                    "Check [%d] %s app launcher processes found running",
+                    expectedCount, Optional.ofNullable(launcherName).map(
+                            str -> "[" + str + "]").orElse("<main>")));
+        } finally {
+            if (pids.length != 0) {
+                killProcess(pids[0]);
+            }
+        }
+    }
+
+    private static long[] findAppLauncherPIDs(JPackageCommand cmd, String launcherName) {
+        // Get the list of PIDs and PPIDs of app launcher processes. Run setWinRunWithEnglishOutput(true) for JDK-8344275.
+        // wmic process where (name = "foo.exe") get ProcessID,ParentProcessID
+        List<String> output = Executor.of("wmic", "process", "where", "(name",
+                "=",
+                "\"" + cmd.appLauncherPath(launcherName).getFileName().toString() + "\"",
+                ")", "get", "ProcessID,ParentProcessID").dumpOutput(true).saveOutput().
+                setWinRunWithEnglishOutput(true).executeAndGetOutput();
+        if ("No Instance(s) Available.".equals(output.getFirst().trim())) {
+            return new long[0];
+        }
+
+        String[] headers = Stream.of(output.getFirst().split("\\s+", 2)).map(
+                String::trim).map(String::toLowerCase).toArray(String[]::new);
+        Pattern pattern;
+        if (headers[0].equals("parentprocessid") && headers[1].equals(
+                "processid")) {
+            pattern = Pattern.compile("^(?<ppid>\\d+)\\s+(?<pid>\\d+)\\s+$");
+        } else if (headers[1].equals("parentprocessid") && headers[0].equals(
+                "processid")) {
+            pattern = Pattern.compile("^(?<pid>\\d+)\\s+(?<ppid>\\d+)\\s+$");
+        } else {
+            throw new RuntimeException(
+                    "Unrecognizable output of \'wmic process\' command");
+        }
+
+        List<long[]> processes = output.stream().skip(1).map(line -> {
+            Matcher m = pattern.matcher(line);
+            long[] pids = null;
+            if (m.matches()) {
+                pids = new long[]{Long.parseLong(m.group("pid")), Long.
+                    parseLong(m.group("ppid"))};
+            }
+            return pids;
+        }).filter(Objects::nonNull).toList();
+
+        switch (processes.size()) {
+            case 2 -> {
+                final long parentPID;
+                final long childPID;
+                if (processes.get(0)[0] == processes.get(1)[1]) {
+                    parentPID = processes.get(0)[0];
+                    childPID = processes.get(1)[0];
+                } else if (processes.get(1)[0] == processes.get(0)[1]) {
+                    parentPID = processes.get(1)[0];
+                    childPID = processes.get(0)[0];
+                } else {
+                    TKit.assertUnexpected("App launcher processes unrelated");
+                    return null; // Unreachable
+                }
+                return new long[]{parentPID, childPID};
+            }
+            case 1 -> {
+                return new long[]{processes.get(0)[0]};
+            }
+            default -> {
+                TKit.assertUnexpected(String.format(
+                        "Unexpected number of running processes [%d]",
+                        processes.size()));
+                return null; // Unreachable
+            }
+        }
+    }
+
     private static boolean isUserLocalInstall(JPackageCommand cmd) {
         return cmd.hasArgument("--win-per-user-install");
     }
 
+    private static boolean isPathTooLong(Path path) {
+        return path.toString().length() > WIN_MAX_PATH;
+    }
+
     private static class DesktopIntegrationVerifier {
 
         DesktopIntegrationVerifier(JPackageCommand cmd, String launcherName) {
@@ -414,19 +577,47 @@ public class WindowsHelper {
         return value;
     }
 
+    private static final class ShortPathUtils {
+        private ShortPathUtils() {
+            try {
+                var shortPathUtilsClass = Class.forName("jdk.jpackage.internal.ShortPathUtils");
+
+                getShortPathWrapper = shortPathUtilsClass.getDeclaredMethod(
+                        "getShortPathWrapper", String.class);
+                // Note: this reflection call requires
+                // --add-opens jdk.jpackage/jdk.jpackage.internal=ALL-UNNAMED
+                getShortPathWrapper.setAccessible(true);
+            } catch (ClassNotFoundException | NoSuchMethodException
+                    | SecurityException ex) {
+                throw rethrowUnchecked(ex);
+            }
+        }
+
+        static Path toShortPath(Path path) {
+            return Path.of(toSupplier(() -> (String) INSTANCE.getShortPathWrapper.invoke(
+                    null, path.toString())).get());
+        }
+
+        private final Method getShortPathWrapper;
+
+        private static final ShortPathUtils INSTANCE = new ShortPathUtils();
+    }
+
     static final Set<Path> CRITICAL_RUNTIME_FILES = Set.of(Path.of(
             "bin\\server\\jvm.dll"));
 
     // jtreg resets %ProgramFiles% environment variable by some reason.
-    private final static Path PROGRAM_FILES = Path.of(Optional.ofNullable(
+    private static final Path PROGRAM_FILES = Path.of(Optional.ofNullable(
             System.getenv("ProgramFiles")).orElse("C:\\Program Files"));
 
-    private final static Path USER_LOCAL = Path.of(System.getProperty(
+    private static final Path USER_LOCAL = Path.of(System.getProperty(
             "user.home"),
             "AppData", "Local");
 
-    private final static String SYSTEM_SHELL_FOLDERS_REGKEY = "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
-    private final static String USER_SHELL_FOLDERS_REGKEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
+    private static final String SYSTEM_SHELL_FOLDERS_REGKEY = "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
+    private static final String USER_SHELL_FOLDERS_REGKEY = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";
 
     private static final Map<String, String> REGISTRY_VALUES = new HashMap<>();
+
+    private static final int WIN_MAX_PATH = 260;
 }
